package FDEV;
#$Id: FDEV.pm,v 1.95 2016/12/24 07:05:22 sofc Exp $

###################################################################
###
###
###      Package FDEV
###
###
###
### The FDEV module contains the base class information and member
### functions for all frequency devices (both potentiostats and 
### frequency response analysers) as well as the FRA and PSTAT classes.
### The FRA and PSTST class both inherit from FDEV and extends this
### for frequency response analysers and potentiostats respectively.
### (they are described separately below).
###
### --\subsection{FDEV}
###
### To obtain a default frequency device instance, call the new function:
### $id = FDEV->new();
### All frequency device instances (both default and real) has the 
### following member functions. All functions are described in details
### where the functions are defined in the file 'FDEV.pm'.
### 
### $id->init()               Initialises the device.
###                           Must always return a true value (usually a 
###                           string identifying the device)
### $id->reset()              Resets the device.
### $id->set_resistor($)      Sets the measure resistor value.
###                           Some devices may not allow changing the resistor.
### $id->get_resistor()       Returns the measure resistor value.
### $id->set_polarisation($$) Sets the DC polarisation of the device to
###                           the specified value (2'nd agument is a setup 
###                           hash reference).
### $id->setup_impedance($)   Setup the device for impedance measurements.
### $id->save()               Reads device configuration and returns list
###                           with setup values. The list can be used directly
###                           with bulk_load_check() and bulk_load().
### $id->checkcmd($)          Sends a cmmand to device and checks for errors
### $id->get_error()          Returns any error from device.
### $id->last_error()         Returns last saved error.
### $id->lase_warning()       Returns last saved warning.
### $id->errorcode($)         Returns the error string correspoding to 
###                           specified code.
### $id->warningcode($)       Returns the warning string correspoding to 
###                           specified code.
### $id->reset_data()         Empties the data array,
### $id->get_data()           Returns the data in the data array.
### $id->virtual()            Returns 1 if a device is a virtual (default)
###                           device.
### $id->bulk_load_check(@)   Loads a list of commands using checkcmd().
### $id->bulk_load(@)         Loads a list of commands using writecmd().
### $id->last_pol_offset()    Returns the last reported polarisation offset.
### $id->setuptime()          Returns the expected setuptime for the device.
### $id->AC_gain()            Returns the AC gain of the device. Default is 1
###                           but some potentiostat devices allow for other 
###                           values.
### $id->get_reference_potentials()
###                           Returns a list of possible reference potentials
###                           Default is 'Fixed' (corresponds to device ground).
### $id->get_default($)       Returns the default setting for the specified 
###                           configuration tag.
### $id->setup_tags()         Returns a list of setup tags which the device
###                           understands.
### $id->setup_tags_chrono()  Returns a list of setup tags used for chrono 
###                           measuremnts.
### $id->frequency_segment_tags()
###                           Returns a list of tags whose value may depend on
###                           frewuency settings and thus change between frequency
###                           sebments
### $id->tag_value($$)        Sets or gets a value for a specific tag. Only valid
###                           for tags listed in frequency_segment_tags()
### $id->tag_description($)   Returns a string describing the tag in question.
### $id->sub setup_tags_potsweep()
###                           Returns a list of setup tags used for cyclic 
###                           voltammetry.
### $id->get_tag_values($)    Returns a list of possible values for a specific
###                           setup key. Is only valid for tags with discrete values.
###                           (enumerators), returns undef for tags with nummeric values
### $id->get_setup_value($$$) Returns the setup value for the specified setup key
###                           See detailed description below!
### $id->socket_client()      TCP-IP socket client for communicating with device.
### $id->set_debugstream($)   Sets the debug stream. If argument is a string, a 
###                           file name is assumed, otherwise a file handle is 
###                           assumed!
### $id->set_errorstream($)   Same as above, but for errors.
### $id->writedebug($)        Writes the argument to the debug stream.
### $id->errorlog($)          Writes the argument to the error srream.
### $id->push_cmd($)          Pushes the argument to the circular command buffer.
### $id->writecmd($)          Writes the argument to the device (using serial_client())
### $id->readcmd()            Returns the result of last command send to the device.
### $id->debugon()            Turns debug on.
### $id->debugoff()           Turns debug off.
### $id->port                 Sets or gets the port number for communication.
### $id->host                 Sets or gets the host name for cummunication.
### $id->delay                Sets or gets the device delay (software only).
### $id->set_mode($)          Sets the device mode (software only, the actual device
###                           mode is usually set as part of set_polarisation()).
### $id->get_modes()          Returns a list of possible modes for the device.
### $id->mode()               Returns the current device mode.
### $id->type()               Returns the device type (name).
### $id->get_minfrequency()   Returns the minimum frequency for the device.
### $id->get_maxfrequency()   Returns the maximum frequency for the device.
### $id->get_minpolarisation()
### $id->get_maxpolarisation()
###                           Returns the minimum and maximum polarisation possible.
### $id->print_setup()        Prints the contents of the current device data 
###                           (software only)
### $id->isalive()            Returns 1 if the device is allive and able to 
###                           communicate 0 otherwise
###
###
### --\subsection{FRA}
### ----\label{module_FRA}
###
###
### The FRA class inherits from FDEV and in addition to the member functions defined 
### above adds the following functions: 
### In some cases it also overloads functions defined in the FDEV class.
### 
### $id->run_frequency($$)    Runs a single frequency measurement (see belov for details)
### $id->run_frequency_all($$)
###                           Same as above, but reports all possible values (which may
###                           include values which makes no sense in the current context).
### $id->get_data_long()      Returns the values in the logndata array.
### $id->get_max_segments()   Returns the maximum number of frequency segnments for the 
###                           device.
### $id->get_max_amplitude()  Returns maximum amplitude allowed for device.
### $id->get_min_amplitude()  Returns minimum amplitude allowed for device.
### $id->amplitude            sets or gets the device AC amplitude (RMS value!).
### $id->get_impedance_time($)
###                           Returns the expected time for impedance for the specified
###                           frequency list (se below for details!).
### $id->get_source_modes()   Returns a list of possible source modes.
### $id->get_imtegration_time($)
###                           Returns the expected time for impedance for the specified
###                           frequency (se below for details!).
### $id->get_min_integration_time($)
###                           Returns the minimum possible integration time for the 
###                           specified frequency (default minimum is 1/f).
###
###
### --\subsection{PSTAT}
### ----\label{module_PSTAT}
###
### The PSTAT class also inherits from FDEV and adds the following member functions:
### In some cases it also overloads functions defined in the FDEV class.
###
### $id->ignore()             Returns the status of the ignore file (used to check if 
###                           another program is using the deivce).
### $id->set_ignore()         Sets the ignore file (creates the file).
### $id->unset_ignore()       Unsets (deletes) the ignore file.
### $id->get_last_OCV()       Returns the value of the last OCV measurement.
### $id->get_minsweeprate()
### $id->get_maxsweeprate()   
###                           Returns the minimum and maximum sweep rates possible.
### $id->start_sweep()        Starts a cyclic voltammetry sweep, returns once the sweep is 
###                           started (after delay)
### $id->sweep_time($)        Returns the expected vlotammetry sweep time based on specified.
###                           setup (a hash ref).
### $id->setup_sweep($)       Setup the device for cyclic voltammetry sweep.
### $id->start_chrono()       Starts a chrono measuremet.
### $id->setup_chrono($)      Setup the device for chrono measurements based on specified
###                           setup (a hash ref).
### $id->run_dvm()            make a single measuremnt of current and vultage and return result.
### $id->max_sweep_points()   Returns the maximum independent potential poitns possible
###                           for the device.
### $id->setup_square($)      Setup square wawe output (usefull for filter/bandwidth selection)
### $id->start_square()       Starts a square wawe function.
### $id->square_time()        returns the time for a square function run
###
###################################################################

use strict;
use warnings;
use Time::HiRes qw(sleep time);
use Fcntl qw(:DEFAULT :flock);
use POSIX qw(/ceil/);
use SocketClient qw();
use Class::Inspector;
our $AUTOLOAD;
our $DEBUGFILE = '/var/log/Frequency_device_communication.log';
our $ERRORFILE = '/var/log/Frequency_device_error.log';
our $HOST = 'localhost';
our $PORT = 12345;
our $err = '';
### Default host and port is for the gpibserver (12345) ###
### on localhost                                        ###

BEGIN {
    use Exporter ();
    our ($VERSION, @ISA, @EXPORT, @EXPORT_OK, %EXPORT_TAGS);
    $VERSION = 1.0;
    %EXPORT_TAGS = (
        'standard'=>[],
		    );
    Exporter::export_ok_tags('standard');
    Exporter::export_tags('standard');
    @ISA = qw(Exporter);
}


sub new
{
    my $class = shift;
    my $self = {
	'_type' => 'NONE',
	'_virtual' => 1,
	'_mode' => undef,
	'_modes' => {'Volt' => 'V','Amp' => 'A'},
	'_minf' => undef,
	'_maxf' => undef,
	'_minpV' => undef,
	'_maxpV' => undef,
	'_minpA' => undef,
	'_maxpA' => undef,
	'_setup' => undef,
	'_default' => {},
	'_desc' => {},
	'_setup_tags' => undef,
	'_setup_tags_chrono' => undef,
	'_setup_tags_potsweep' => undef,
	'_port' => $PORT,
	'_host' => $HOST,
	'_addr' => 0,
	'_debug' => 0,
	'_overhead' => 1,
	'_AC_gain' => 1,
	'_resistor' => 1,
	'_values_resistor' => [1],
	'_errorstream' => $ERRORFILE,
	'_debugstream' => $DEBUGFILE,
	'_lastcmd' => [],
	'_last_error' => '',
	'_last_warning' => '',
	'_errorcodes' => {
	    '' => ' ',
	    'XX' => 'No data returned',
	},
	'_warningcodes' => {'' => ' '},
	'_save_tags' => [],
	'_values_reference_potential' => ['Fixed'],
	'_setuptime' => 1,
	'_delay' => 0,
	'_data' => [],
	'_pol_offset' => 0,
	'_freq_seg_tags' => [],
	'_amplitude' => 0,
    };
    bless $self,$class;
    return $self;
}

### Stub methods to be overridden by individual devices ###

sub init()
{
    #################################################
    ### init()                                    ###
    ### Initialises the device and clears errors  ###
    ### as well as performs some basic setup      ###
    ### (beep off etc.)                           ###
    ### Returns empty string or device identifier ###
    #################################################
    return 'NONE';
}

sub reset()
{
    #################################################
    ### reset()                                   ###
    ### Resets the device, may not be the same as ###
    ### init(), but can be for some devices!      ###
    ### Gnerally a reset() command will release   ###
    ### generator output (AC or DC or both,       ###
    ### depends on device type)!                  ###
    #################################################
    return undef;
}

sub set_resistor($)
{
    ########################################
    ### set_resistor($)                  ###
    ### Set the shunt ressitor value     ###
    ### Returns the last set value for   ###
    ### the internal current shunt       ###
    ### resistor, default is 1           ###
    ########################################
    my $self = shift;
    return $self->{'_resistor'};
}

sub set_polarisation($$)
{
    #################################################
    ### set_polarisation($$)                      ###
    ### Setup the DC polarisation of the device   ###
    ### to the specified value, returns the       ###
    ### returns the polarisation value set which  ###
    ### may be different than the value specified ###
    ### if the value is outside the legal range   ###
    ### for example (other reasons may also be    ###
    ### possible). The second argument must be    ###
    ### a reference to a setup hash               ###
    ##################################################################
    ### If the setup key 'polarisation_offset' is set (and is      ###
    ### a function pointer) the function is called and the result  ###
    ### is saved in the '_pol_offset' key and added to the         ###
    ### specified polarisation!                                    ###
    ##################################################################
    my $self = shift;
    my $pol = shift;
    my $setup = shift;
    return 0;
}

sub setup_impedance($)
{
    ##################################################################
    ### setup_impedance($)                                         ###
    ### setup a device for frequency response analysis             ###
    ### (physically possible also for a potentiostat/galvanostat   ###
    ### in conjunction with a frequency response analyser device!) ###
    ### returns the DC polarisation of the device (which may be    ###
    ### different than the one specified in the 'polarisation' key ###
    ### of the supplied setup hash reference                       ###
    ### If the setup key 'polarisation_offset' is set (and is      ###
    ### a function pointer) the function is called and the result  ###
    ### is added to the polarisation                               ###
    ### This is usually done by using the set_polarisation method  ###
    ##################################################################
    my $self = shift;
    my $setup = shift;
    return 0;
}

sub save()
{
    ##################################################
    ### save()                                     ###
    ### Reads the values corresponding to the tags ###
    ### given in the '_save_tags' keys. Returns an ###
    ### array containing command sequences which   ###
    ### can be supplied individually to writecmd() ###
    ### or the entire array to bulk_load()         ###
    ##################################################
    my $self = shift;
    return ();
}

sub checkcmd(@)
{   
    ################################################
    ### checkcmd(@)                              ###
    ### Sends the specified commands to the      ###
    ### device using default tcp/IP socket comms ###
    ### Only tries to write if address is not 0  ###
    ### Unlike writecmd() this function tries to ###
    ### detect errors and redo the command if    ###
    ### errors are encounted. Note not all devs  ###
    ### supports thic function, thus default is  ###
    ### to call writecmd()                       ###
    ################################################
    my $self = shift;
    return $self->writecmd(@_);
}


sub get_error()
{
    ###############################################
    ### get_error()                              ##
    ### Returns any error for the debvice,       ##
    ### Either as a string or a errorcode        ##
    ###############################################
    my $self = shift;
    return '';
}

sub isallive()
{
    ###############################################
    ### isallive()                               ##
    ### Returns 1 if the device is able to       ##
    ### communicate, 0 otherwise, the default is ##
    ### 1                                        ##
    ###############################################
    my $self = shift;
    return 1;
}

### End stubs ###

sub last_error()
{ 
    ###############################################
    ### last_error()                            ###
    ### Returns last error reported by device   ###
    ### Note this only includes measurement     ###
    ### errors and not communication errrors!   ###
    ###############################################
    my $self = shift;
    return $self->{'_last_error'};
}

sub last_warning()
{
    ###############################################
    ### last_warning()                          ###
    ### Returns last warning reported by device ###
    ###############################################
    my $self = shift;
    return $self->{'_last_warning'};
}

sub errorcode($)
{
    ###############################################
    ## errorcode($)                              ##
    ## returns a 'human readable' description of ##
    ## the errorcode specified. If no string is  ##
    ## defined, the cerrorcode is simply         ##
    ## returned as is                            ##
    ###############################################
    my $self = shift;
    my $error = shift;
    return $self->{'_errorcodes'}{$error} if (exists($self->{'_errorcodes'}{$error}));
    return "Unknown error: '$error'";
}

sub warningcode($)
{
    ###############################################
    ## warningcode($)                            ##
    ## returns a 'human readable' description of ##
    ## the warningcode specified. If no string   ##
    ## is defined, the warningcode is simply     ##
    ## returned as is                            ##
    ###############################################
    my $self = shift;
    my $warning = shift;
    return $self->{'_warningcodes'}{$warning} if (exists($self->{'_warningcodes'}{$warning}));
    return "Unknown warning: '$warning'";
}


sub reset_data()
{
    ###############################
    ### reset_data()            ###
    ### Deletes any data in the ###
    ### '_data' array           ###
    ###############################
    my $self = shift;
    $self->{'_data'} = [];
}

sub get_data()
{
    #################################
    ### get_data()                ###
    ### returns the '_data' array ###
    #################################
    my $self = shift;
    return @{$self->{'_data'}};
}

sub virtual()
{
    ###################################################
    ### virtual()                                   ###
    ### returns if the device is a real one or not  ###
    ### must return 1 for virtual devices and 0 for ###
    ### real devices!                               ###
    ###################################################
    my $self = shift;
    return $self->{'_virtual'};
}

sub bulk_load_check(@)
{
    ##################################################
    ### bulk_load_check()                          ###
    ### Calls checkcmd() for each of the arguments ###
    ### Usefull for bulk loading lots of commands  ###
    ### unlike bulk_load, this function checks for ###
    ### errors if checkcmd() is defined to do so!  ###
    ##################################################
    my $self = shift;
    while (@_)
    {
	my $cmd = shift;
	$self->checkcmd($cmd);
    }
    return undef;
}

sub bulk_load(@)
{
    ##################################################
    ### bulk_load()                                ###
    ### Calls checkcmd() for each of the arguments ###
    ### Usefull for bulk loading lots of commands  ###
    ### Note, does not check for errors!           ###
    ##################################################
    my $self = shift;
    while (@_)
    {
	my $cmd = shift;
	$self->writecmd($cmd);
    }
    return undef;
}

sub last_pol_offset()
{
    ##################################################
    ### last_pol_offset()                          ###
    ### Returns the last value of any remote       ###
    ### polarisation offset function call          ###
    ##################################################
    my $self = shift;
    return $self->{'_pol_offset'};
}

sub setuptime()
{
    ################################################
    ### setuptime()                              ###
    ### Returns the expected stupt time for the  ###
    ### device. Default is 1 second              ###
    ################################################
    my $self = shift;
    return $self->{'_setuptime'};
}

sub AC_gain()
{
    ################################################
    ### AC_gain()                                ###
    ### Returns the AC gain for the device,      ###
    ### dfault is 1, but some instruments can    ###
    ### have other values (in some cases this    ###
    ### can be specified). The value returned is ###
    ### the difference between the ratio between ###
    ### the output and input or specified value  ###
    ### (for potentiostats and FRA respectively) ###
    ### NB value NOT in db!                      ###
    ################################################
    my $self = shift;
    return $self->{'_AC_gain'};
}

sub get_reference_potentials()
{
    #################################################
    ### get_reference_potentials()                ###
    ### returns the possible reference potneitals ###
    ### Default is 'Fixed'. Some devices also     ###
    ### supports the 'OCV' mode where the applied ###
    ### potential is added the the rest potential ###
    ### of the sample under test                  ###
    #################################################
    my $self = shift;
    return @{$self->{'_values_reference_potential'}};
}

sub get_default()
{
    ##############################################
    ### get_default()                          ###
    ### returns the default setting for the    ###
    ### tag specified, if no tag is specified  ###
    ### the complete default hash is returned  ###
    ##############################################
    my $self = shift;
    my $tag = shift;
    if ($tag)
    {
	return $self->{'_default'}{$tag} if (exists($self->{'_default'}{$tag}));
	return undef;
    }
    return %{$self->{'_default'}};
}


sub setup_tags()
{
    ###############################################
    ### setup_tags()                            ###
    ### Returns a list of acceptable setup tags ###
    ### Can be used by a application to easialy ###
    ### get the setup tags and values           ###
    ###############################################
    my $self = shift;
    my @res = ();
    @res = split(',',$self->{'_setup_tags'}) if ($self->{'_setup_tags'});
    return @res;
}

sub tag_description($)
{
    ################################################
    ### Returns a text string describing the tag ###
    ### in question, returns empty string if no  ###
    ### info is found                            ###
    ################################################
    my $self = shift;
    my $key = shift;
    return $self->{'_desc'}{$key} if (exists($self->{'_desc'}{$key}));
    return '';
}


sub setup_tags_chrono()
{
    ###############################################
    ### setup_tags()                            ###
    ### Returns a list of acceptable setup tags ###
    ### Can be used by a application to easialy ###
    ### get the setup tags and values           ###
    ###############################################
    my $self = shift;
    my @res = ();
    @res = split(',',$self->{'_setup_tags_chrono'}) if ($self->{'_setup_tags_chrono'});
    return @res;
}

sub setup_tags_potsweep()
{
    ###############################################
    ### setup_tags()                            ###
    ### Returns a list of acceptable setup tags ###
    ### Can be used by a application to easialy ###
    ### get the setup tags and values           ###
    ###############################################
    my $self = shift;
    my @res = ();
    @res = split(',',$self->{'_setup_tags_potsweep'}) if ($self->{'_setup_tags_potsweep'});
    return @res;
}

sub get_tag_values($)
{
    ##############################################
    ### get_tag_values($)                      ###
    ### returns a list of the possible values  ###
    ### the specified tag can have.            ###
    ### Returns undef if not defined which is  ###
    ### the case if the tag can be any value   ###
    ### (in the case of floating point values  ###
    ### for instance)                          ###
    ### This method (and the corresponding     ###
    ### values) are mainly to be used for UI   ###
    ### purposes                               ###
    ##############################################
    my $self = shift;
    my $tag = '_values_'.shift;
    return undef unless (exists($self->{$tag}));
    my @res = ();
    @res = @{$self->{$tag}} if ($self->{$tag});
    return @res;
}

sub frequency_segment_tags()
{
    ###############################################
    ### frequency_segment_tags()                ###
    ### Returns a list of tags whose value may  ###
    ### have to be changed between individual   ###
    ### frequency segments. All tags defined by ###
    ### this function must be valid input for   ###
    ### tag_value($$)                           ###
    ###############################################
    my $self = shift;
    return @{$self->{'_freq_seg_tags'}};
}

sub tag_value
{
    ####################################################
    ### tag_value($)                                 ###
    ### Gets of sets the value for the specified tag ###
    ### The tags defined inn '_freq_seg_tags' must   ###
    ### be supproted by this function                ###
    ####################################################
    my $self = shift;
    my $tag = shift;
    return undef unless (@_);
    my $val = shift;
    return $val;
}


sub get_setup_value($$$)
{
    ##############################################################
    ### get_setup_value($$$)                                   ###
    ### This fnction checks the specified setup hashref for    ###
    ### the given key and if it exists returns the value.      ###
    ### If the key is not found, the default is returned       ###
    ### The function bails with an error if the requested      ###
    ### key is not contained in the list given by setup_tags() ###
    ### setup_tags_chrono() or setup_tags_potsweep()           ###
    ### (Usefull for internal consistency checks)              ###
    ##############################################################
    my $self = shift;
    my $default = shift;
    my $setup = shift;
    my $tag = shift;
    my $found = 0;
    foreach ($self->setup_tags())
    {
	$found = 1 if ($_ eq $tag);
    }
    foreach ($self->setup_tags_chrono())
    {
	$found = 1 if ($_ eq $tag);
    }
    foreach ($self->setup_tags_potsweep())
    {
	$found = 1 if ($_ eq $tag);
    }
    die "Tag $tag not found in setup list for device $self!" unless ($found);
    $default = $$setup{$tag} if (exists($$setup{$tag}) && defined($$setup{$tag}));
    return $default;
}

sub socket_client(@)
{
    my $self = shift;
    my $port = $self->{'_port'};
    my $remote = $self->{'_host'};
    return SocketClient::socket_client($remote,$port,@_);
}

sub set_debugstream($)
{
    #########################################################
    ### set_debugstream($)                                ###
    ### Sets the debugstream to a new value               ###
    #########################################################
    my $self = shift;
    $self->{'_debugstream'} = shift;
}

sub set_errorstream($)
{
    #########################################################
    ### set_errorstream($)                                ###
    ### Sets the errorstream to a new value               ###
    #########################################################
    my $self = shift;
    $self->{'_errorstream'} = shift;
}

sub writedebug($)
{
    #############################################################
    ### writedebug($)                                         ###
    ### Writes the specified string to the debug stream if    ###
    ### _debug is true                                        ###
    #############################################################
    my $self = shift;
    return unless ($self->{'_debug'});
    my $str = shift;
    chomp $str;
    my $time = time . ', '. localtime;
    my $idstr = "Host:$HOST, Port:$self->{'_port'}, Addr:$self->{'_addr'}";
    if (ref($self->{'_debugstream'}) =~ /GLOB/)
    {
	### Assume stream is a file handle which is writable ##
	print($self->{'_debugstream'},"$time, $idstr, String: $str\n") or die "Could not write to handle $self->{'_debugstream'}";	
    }
    else
    {
	sysopen(DFH, $self->{'_debugstream'}, O_WRONLY | O_CREAT) or die "Error: can't open file: $self->{'_debugstream'} $!\n";
	flock(DFH, LOCK_EX);
	open(DEB,'>>',$self->{'_debugstream'}) or die "Could not open debugfile $self->{'_debugstream'} $!";
	print DEB "$time, $idstr, String: $str\n";
	close(DEB);
	flock(DFH, LOCK_UN);
	close(DFH);
    }
}

sub errorlog($)
{
    #############################################################
    ### errorlog($)                                           ###
    ### Writes the specified string to the error stream       ###
    ### Note that the errorstream is not stderr, but a file!  ###
    #############################################################
    my $self = shift;
    my $str = shift;
    chomp $str;
    my $lastcmd = join(";",@{$self->{'_lastcmd'}});
    my $time = time . ', '. localtime;
    my $idstr = "Host:$HOST, Port:$self->{'_port'}, Addr:$self->{'_addr'}";
    #######################################
    ## Getting complete stack backtrace ###
    ## for anything else than warnings  ###
    #######################################
    my @trace = ();
    my $i = 0;
    while(my ($package,$filename,$line,$subroutine) = caller($i++))
    {
        unshift @trace,"'$filename' line $line called $subroutine";
    }
    my $tstr = ' ### Stack backtrace: '.join(' -> ',@trace);
    if (ref($self->{'_errorstream'}) =~ /GLOB/)
    {
	### Assume that stream is a file handle that we can write to ####
	print($self->{'_errorstream'},"$time, $idstr, Last commands: '$lastcmd', String: $str $tstr\n") or die "Could not write to handle $self->{'_errorstream'}";	
    }
    elsif (-e $self->{'_errorstream'} && -w $self->{'_errorstream'})
    {
	sysopen(DFH, $self->{'_errorstream'}, O_WRONLY | O_CREAT) or die "Error: can't open file: $self->{'_errorstream'} $!\n";
	flock(DFH, LOCK_EX);
	open(DEB,'>>',$self->{'_errorstream'}) or die "Could not open debugfile $self->{'_errorstream'} $!";
	print DEB "$time, $idstr, Last commands: '$lastcmd', String: $str\n";
	close(DEB);
	flock(DFH, LOCK_UN);
	close(DFH);
    }
    else
    {
	warn("$time, $idstr, Last commands: '$lastcmd', String: $str\n");
    }
}


sub push_cmd($)
{
    ###############################################
    ### push_cmd()                              ###
    ### appends the given command to the list   ###
    ### of commands. Only the last 5 commands   ###
    ### are retained                            ###
    ###############################################
    my $self = shift;
    my $cmd = shift;
    my @cmds = ();
    @cmds = @{$self->{'_lastcmd'}} if ($self->{'_lastcmd'});
    push @cmds,$cmd;
    while (@cmds > 5)
    {
	shift @cmds;
    }
    $self->{'_lastcmd'} = \@cmds;
}

sub writecmd(@)
{   
    ################################################
    ### writecmd(@)                              ###
    ### Sends the specified commands to the      ###
    ### using default tcp/IP socket comms        ###
    ### Only tries to write if address is not 0  ### 
    ################################################
    my $self = shift;
    return  unless ($self->{'_addr'});
    $self->push_cmd(join("\t",@_));
    $self->writedebug("WRITE : ".join("\t",@_));
    my $res = $self->socket_client('w',$self->{'_addr'},@_);
    return if defined($res);
    $self->errorlog("WRITE Failed!: $err");
    $err = '';
    return undef;
}

sub readcmd()
{
    #################################################
    ### readcmd()                                 ###
    ### sends a read request to the device and    ###
    ### returns the string returned by the device ###
    ### Only tries to read if address is not 0    ### 
    #################################################
    my $self = shift;
    return '' unless ($self->{'_addr'});
    my $res = $self->socket_client('r',$self->{'_addr'});
    if (defined($res))
    {
	chomp $res;
	chomp $res;
	$self->writedebug("READ : ".$res);
    }
    else
    {
	$self->writedebug("READ : <NULL>");
    }
    return $res if defined($res);
    $self->errorlog("READ Failed!: $err");
    $err = '';
    return undef;
}

sub debugon()
{
    ###########################
    ### debugon()           ###
    ### Turns on debugging  ###
    ###########################
    my $self = shift;
    $self->{'_debug'} = 1;
}

sub debugoff()
{
    ###########################
    ### debugoff()          ###
    ### Turns off debugging ###
    ###########################
    my $self = shift;
    $self->{'_debug'} = 0;
}

sub port
{
    #############################################
    ### port([opt $port])                     ###
    ### Used without argument returns the     ###
    ### current port number, used with arguemt###
    ### sets the device communication port to ###
    ### the specified value (default = 12345) ###
    #############################################
    my $self = shift;
    if (@_)
    {
	$self->{'_port'} = shift;
    }
    return $self->{'_port'};
}

sub host
{
    #################################################
    ### host([opt $host])                         ###
    ### Used without argument returns the         ###
    ### current host, used with arguemt           ###
    ### sets the device communication host to     ###
    ### the specified value (default = localhost) ###
    #################################################
    my $self = shift;
    if (@_)
    {
	$self->{'_host'} = shift;
    }
    return $self->{'_host'};
}

sub delay
{
    #####################################
    ### delay                         ###
    ### gets or sets the device delay ###
    #####################################
    my $self = shift;
    if (@_)
    {
	$self->{'_delay'} = shift;
    }
    return $self->{'_delay'};
}


sub set_mode($)
{
    ######################################
    ### set_mode($)                    ###
    ### sets the device mode to either ###
    ### voltage or current (V or A)    ###
    ### Note some devices may support  ###
    ### more as well as less modes     ###
    ### than described                 ###
    ######################################
    my ($self ,$mode) = @_;
    $self->{'_mode'} = undef;
    $self->{'_mode'} = $self->{'_modes'}->{$mode} if exists ($self->{'_modes'}->{$mode});
    die "ERROR: Illegal mode '$mode' attempted for $self->{'_type'}!" unless ($self->{'_mode'});
    return $self;
}

sub get_modes()
{
    #####################################
    ### get_modes()                   ###
    ### returns a list of valid modes ###
    ### for the device                ###
    #####################################
    my $self = shift;
    return (sort keys %{$self->{'_modes'}})
}

sub mode()
{
    ######################################
    ### mode()                         ###
    ### returns the device mode        ###
    ######################################
    my $self = shift;
    return $self->{'_mode'};
}

sub type()
{
    ######################################
    ### type()                         ###
    ### returns the device type (name) ###
    ######################################
    my $self = shift;
    return $self->{'_type'};
}

sub get_minfreqency()
{
    ######################################
    ### get_minfrequency()             ###
    ### returns the minimum frequency  ###
    ### for the device                 ###
    ######################################
    my $self = shift;
    return $self->{'_minf'};
}
sub get_maxfreqency()
{
    ######################################
    ### get_maxfrequency()             ###
    ### returns the maximum frequency  ###
    ### for the device                 ###
    ######################################
    my $self = shift;
    return $self->{'_maxf'};
}

sub get_minpolarisation()
{
    #########################################
    ### get_minpolarisation()             ###
    ### returns the minimum polarisation  ###
    ### for the device                    ###
    #########################################
    my $self = shift;
    if ($self->mode eq 'A')
    {
	return $self->{'_minpA'};
    }
    else
    {
	return $self->{'_minpV'};
    }
}
sub get_maxpolarisation()
{
    #########################################
    ### get_maxpolarisation()             ###
    ### returns the maximum polarisation  ###
    ### for the device                    ###
    #########################################
    my $self = shift;
    if ($self->mode eq 'A')
    {
	return $self->{'_maxpA'};
    }
    else
    {
	return $self->{'_maxpV'};
    }
}

sub AUTOLOAD()
{
    ################################################################
    ### AUTOLOAD()                                               ###
    ### If caller ended here, undefiend subroutine call happened ###
    ################################################################
    my $self = shift;
    my $type = ref ($self) or die "$self is not an object!";
    die "Unknown function call $AUTOLOAD attempted!";
}

sub DESTROY()
{

}

sub print_setup()
{
    #############################################
    ### print_setup()                         ###
    ### utility function for debugging        ###
    ### purposes, prints the contents of the  ###
    ### device instance data fields $self->{} ###
    #############################################
    my $self = shift;
    my @arr = ('*******************************************');
    push @arr,'Device ID: '.$self->{'_type'};
    push @arr,'Member functions:';
    my $class = $self;
    $class =~ s/=.*//;
    my @methods = Class::Inspector->methods( $class, 'full', 'public' );
    foreach my $m (@methods)
    {
	if (ref($m) =~ /ARRAY/)
	{
	    foreach (@{$m})
	    {
		push @arr,"\t$_";
	    }
	    next;
	}
	push @arr,"\t$m";
    }
    push @arr,'';
    push @arr,'Instance data (Field,  Value):';
    foreach my $key (keys %{$self})
    {
	if (defined ($self->{$key}))
	{
	    if (ref($self->{$key}) =~ /ARRAY/)
	    {
		push @arr,"$key, Array with values:";
		push @arr,"\tIndex\tValue";
		my $x = 0;
		foreach (@{$self->{$key}})
		{
		    push @arr,"\t$x\t'$_'";
		    $x++;
		}
		push @arr,'';
		next;
	    }
	    if (ref($self->{$key}) =~ /HASH/)
	    {
		push @arr,"$key, Asociative array with values:";
		push @arr,"\tKey\tValue";
		foreach my $k (keys %{$self->{$key}})
		{
		    push @arr,"\t'$k'\t'$self->{$key}{$k}'";
		}
		push @arr,'';
		next;
	    }	    
	    push @arr,$key.",\t'$self->{$key}'"; 
	}
	else
	{
	    push @arr,$key.",\t\\N"; 
	}
    }
    push @arr,'';
    return join("\n",@arr);
}

sub get_resistor()
{
    ########################################
    ### get_resistor()                   ###
    ### Returns the last set value for   ###
    ### the internal current shunt       ###
    ### resistor, default is 1           ###
    ########################################
    my $self = shift;
    return $self->{'_resistor'};
}


### FRA extends FDEV ###

package FRA;
our @ISA = qw(FDEV);

sub new
{
    my $class = shift;
    my $self = $class->SUPER::new();
    $self->{'_minampV'} = 0;
    $self->{'_minampA'} = 0;
    $self->{'_maxampV'} = undef;
    $self->{'_maxampA'} = undef;
    $self->{'_minintegraton'} = 0;
    $self->{'_source_modes'} = {},
    $self->{'_values_frequency_order'} = ['normal','reverse','random'];
    $self->{'_addr'} = shift;
    $self->{'_setup'} = shift if (@_);
    $self->{'_function_reload'} = 5;
    $self->{'_delay'} = 0;
    $self->{'_longdata'} = [];
    $self->{'amplitude'} = 0;
    ## Default no practical limit on frequency segments ##  
    $self->{'_max_frequency_segments'} = 100;  
    die "No address specified!" unless defined $self->{'_addr'};
    bless $self,$class;
    return $self;
}

### Stub method to be overridden by individual device ###
sub run_frequency_all($$)
{
    ###############################################################
    ### run_frequency_all($$)                                   ###
    ### runs a single frequency measurement as run_frequency()  ###
    ### unlike the normal function, this function returns as    ###
    ### many data as possible (including the valtages of the    ###
    ### individual inputs if possible)                          ###
    ### Defaults to run_frequency($$) if the device does not    ###
    ### support the additional functionality nescesarry         ###
    ###############################################################
    my $self = shift;
    my $string = $self->run_frequency(@_);
    push @{$self->{'_longdata'}},$string;
    return $string;
}


sub run_frequency($$)
{
    ################################################################
    ### run_frequency($$)                                        ###
    ### runs a single frequency measurement                      ###
    ### First argument is frequency, second argument is the      ###
    ### integration time and third argument (if any) is          ###
    ### any extra setup information in the form af a hash ref    ###
    ### Returns the result of the single frequency measurement   ###
    ### Format of return value depends on setup!                 ###
    ### Default is a space separated string: 'freq real imag'    ###
    ### The default device returns 'freq 1.000000 -1.000000'     ###
    ### as a hard coded value (for debugging/testing purposes)   ###
    ### If the setup hash contains the key 'resistor' then this  ###
    ### value is multiplied onto the real and imaginary values   ###
    ### before returning the result (default is 1)!              ###
    ### If the setup hash contains the entry 'function' then the ###
    ### referenced function is run every 5 seconds durring the   ###
    ### measurement (usefull for plotting purposes). Note the    ###
    ### function is called without any arguments!                ###
    ################################################################
    my $self = shift;
    my $freq = shift;
    my $time = shift;
    my $setup = shift;
    my $testfile = undef;
    $testfile = $$setup{'semafore_file'} if (exists $$setup{'semafore_file'} && $$setup{'semafore_file'});
    my $f = 1;
    $f = $$setup{'resistor'} if (exists($$setup{'resistor'}));
    my $etime = time + $time;
    my $count = 1;
    while (time < $etime)
    {
        sleep(1);
	if (exists($$setup{'function'}) && $count++ > $self->{'_function_reload'})
	{
	    $count = 1;
	    $$setup{'function'}->();
	}
        next if (!$testfile || -e $testfile);
        return undef;
    }
    my $string = sprintf("%f %f %f",$freq,$f,-$f);
    push @{$self->{'_data'}},$string;
    return $string;
}

### End stub methode

sub reset_data()
{
    my $self = shift;
    $self->{'_data'} = [];
    $self->{'_longdata'} = [];
}

sub get_data_long()
{
    #####################################
    ### get_data_long()               ###
    ### returns the '_longdata' array ###
    #####################################
    my $self = shift;
    return @{$self->{'_longdata'}};
}

sub get_max_segments()
{
    ###############################################
    ### get_max_segments()                      ###
    ### returns the maximum number of frequency ###
    ### segments. Default is 100 (for practical ###
    ### purposes unlimited), but some devices   ###
    ### may enforce a lower limit.              ###
    ###############################################
    my $self = shift;
    return $self->{'_max_frequency_segments'};
}

sub amplitude
{
    ##################################################
    ### amplitude($)                               ###
    ### gets or sets the generator amplitude       ###
    ##################################################
    my $self = shift;
    if (@_)
    {
	$self->{'_amplitude'} = shift;
    }
    return $self->{'_amplitude'};
}

sub get_impedance_time($)
{
    ################################################
    ### get_impedance_time($)                    ###
    ### Returns the expected time for a complete ###
    ### impedance spectrum (including setup time ###
    ### overhead and individual integration      ###
    ### times. it accepte either an array of     ###
    ### frequencies or an array-ref              ###
    ### in the case of the array ref, the        ###
    ### content of the array may be either a     ###
    ### a list of frequencies or a hash with     ###
    ### the frequency in the 'frequency' key and ###
    ### the integration time in the 'time' key   ###
    ### If the time is not defined, the function ###
    ### assumes minimum integration time.        ###
    ### The function returns the time in seconds ###
    ################################################
    my $self = shift;
    my $frqlst = shift;
    my $totaltime = $self->{'_setuptime'};
    return $totaltime unless ($frqlst);
    if (ref($frqlst) !~ /ARRAY/)
    {
	$totaltime += $self->get_integration_time($frqlst,0);
	while (@_)
	{
	    my $x = shift;
	    $totaltime += $self->get_integration_time($x,0);
	}
	return POSIX::ceil($totaltime * $self->{'_overhead'});
    }
    foreach my $x (@{$frqlst})
    {
	if (ref($x) =~ /HASH/)
	{
	    if (exists($$x{'time'}) && $$x{'time'})
	    {
		$totaltime += $self->get_integration_time($$x{'frequency'},$$x{'time'});
	    }
	    else
	    {
		$totaltime += $self->get_integration_time($x,0);
	    }
	}
	else
	{
	    $totaltime += $self->get_integration_time($x,0);
	}
    }
    return POSIX::ceil($totaltime * $self->{'_overhead'});
}


sub get_max_amplitude()
{
    #########################################
    ### get_max_amplitude()               ###
    ### returns the maximum amplitude     ###
    ### for the device                    ###
    #########################################
    my $self = shift;
    if ($self->mode eq 'A')
    {
	return $self->{'_maxampA'};
    }
    else
    {
	return $self->{'_maxampV'};
    }
}

sub get_min_amplitude()
{
    #########################################
    ### get_min_amplitude()               ###
    ### returns the minimum amplitude     ###
    ### for the device                    ###
    #########################################
    my $self = shift;
    if ($self->mode eq 'A')
    {
	return $self->{'_minampA'};
    }
    else
    {
	return $self->{'_minampV'};
    }
}

sub get_source_modes()
{
    #####################################
    ### get_modes()                   ###
    ### returns a list of valid modes ###
    ### for the device                ###
    #####################################
    my $self = shift;
    return (sort keys %{$self->{'_source_modes'}})
}

sub get_integration_time($$)
{
    ###############################################################
    ### get_integration_time($f,$time)                          ###
    ### returns the estimated integration time for frequency $f ###
    ### If the $time argument is missing the minimum time is    ###
    ### returned as is the case if the minimum time is larger   ###
    ### than the requested integratiojn time $time              ###
    ### for some devices, the returned time will differ         ###
    ### considerably from the requested time                    ###
    ###############################################################
    my $self = shift;
    my $freq = shift;
    my $time = shift;
    $time += $self->{'_delay'};
    my $mintime = $self->get_min_integration_time($freq);
    return $mintime unless ($time && $time > $mintime);
    return POSIX::ceil($time);
}

sub get_min_integration_time($)
{
    ##############################################################
    ### get_min_integration_time($f)                           ###
    ### returns the minimum integration time for frequency $f  ###
    ##############################################################
    my $self = shift;
    my $freq = shift;
    if ($self->virtual)
    {
	my $time = 1/$freq;
	return POSIX::ceil($time);
    }
    $freq = $self->{'_minf'} if ($freq < $self->{'_minf'});
    $freq = $self->{'_maxf'} if ($freq > $self->{'_maxf'});
    my $time = 1/$freq;
    $time = $time + $self->{'_delay'};
    $time = $self->{'_minintegration'} if ($time < $self->{'_minintegration'});
    return POSIX::ceil($time);
}




#### Potentiostat extends FDEV ####

package PSTAT;
our @ISA = qw(FDEV);

sub new
{
    my $class = shift;
    my $self = $class->SUPER::new();
    $self->{'_maxs'} = undef;
    $self->{'_mins'} = undef;
    $self->{'_addr'} = shift;
    $self->{'_setup'} = shift if (@_);
    $self->{'_last_ocv'} = undef;
    $self->{'_ignorefile'} = '/tmp/PSTAT_ignore';
    die "No address specified!" unless defined $self->{'_addr'};
    bless $self,$class;
    return $self;
}

sub ignore()
{
    #############################################
    ### ignore()                              ###
    ### Returns the status of the ignore file ###
    #############################################
    my $self = shift;
    return -e $self->{'_ignorefile'};
}

sub set_ignore()
{
    #############################################
    ### set_ignore()                          ###
    ### Creates the ignore file (the semafore ###
    ### which tells any chrono measurements   ###
    ### not to do any easurements (for        ###
    ### instance durring an impedance scan)). ###
    #############################################
    my $self = shift;
    return if (-e $self->{'_ignorefile'});
    open MYOUT,'>',$self->{'_ignorefile'} or die "Could not open file $self->{'_ignorefile'} for writing $!";
    close MYOUT;
}

sub unset_ignore()
{
    #############################################
    ### unset_ignore()                        ###
    ### Removes the ignore file (the semafore ###
    ### which tells any chrono measurements   ###
    ### not to do any easurements (for        ###
    ### instance durring an impedance scan)). ###
    #############################################    
    my $self = shift;
    return unless (-e $self->{'_ignorefile'});
    unlink $self->{'_ignorefile'};
}


sub get_last_ocv()
{
    ###########################################
    ### get_last_ocv()                      ###
    ### Returns the last measured OCV value ###
    ### returns undef if device does not    ###
    ### support this or the ocv has not     ###
    ### been measured                       ###
    ###########################################
    my $self = shift;
    return $self->{'_last_ocv'};
}

sub get_minsweeprate()
{
    #########################################
    ### get_minsweeprate()                ###
    ### returns the minimum sweep rate    ###
    ### for the device                    ###
    #########################################
    my $self = shift;
    return $self->{'_mins'};
}
sub get_maxsweeprate()
{
    #########################################
    ### get_maxsweeprate()                ###
    ### returns the maximum sweep rate    ###
    ### for the device                    ###
    #########################################
    my $self = shift;
    return $self->{'_maxs'};
}

### Stub methode to be overridden by individual device ###


sub start_sweep()
{
    ###########################################
    ### start_sweep()                       ###
    ### starts a linear polarisation sweep  ###
    ###########################################
    my $self = shift;
    return undef;
}

sub start_square()
{
    ##########################################
    ### start_square()                     ###
    ### Starts a square wawe output        ###
    ##########################################
    my $self = shift;
    return $self->start_sweep();
}

sub sweep_time($)
{
    ####################################################
    ### sweep_time($$)                               ###
    ### returns the expected sweep time for the      ###
    ### specified setup information (a hash ref)     ###
    ### and the potential point list (an array ref)  ###
    ### the returned time includes the initial delay ###
    ### time as opposed to the time returned by      ###
    ### setup_sweep()                                ###
    ####################################################
    my $self = shift;
    my $setup = shift;
    my $potlist = shift;
    return 0;
}

sub square_time()
{
    my $self = shift;
    my $setup = shift;
    return 0;
}

sub setup_square()
{
    ######################################################
    ### setup_square()                                 ###
    ### Setup a square wawe output with stepheight as  ###
    ### specified. If a second argument is specified,  ###
    ### it must be a setup hashref                     ###
    ### 
    ### Note not all potentiostats may support this    ###
    ### functionality!                                 ###
    ######################################################
    my $self = shift;
    my $stepsize = shift;
    my $setup = shift;
    return (0,0);
}

sub setup_sweep($)
{
    #######################################################
    ### setup_sweep($$)                                 ###
    ### setup a linear polarisation sweep based on      ###
    ### the supplied setup information (a hash ref)     ###
    ### and the potential point list (an array ref)     ###
    ### Returns the true polarisation of the first      ###
    ### point as well as the expected sweep time        ###
    ### Note that the sweep time returned form this     ###
    ### function does NOT include the initial delay     ###
    ### as the start_sweep() function first returns     ###
    ### after the delay has expired (making the time    ###
    ### the actual sweep the one needed).               ###
    ################################################################## 
    ### If the setup key 'polarisation_offset' is set (and is      ###
    ### a function pointer) the function is called and the result  ###
    ### is added to the start polarisation as well as the          ###
    ### individual points in the polarisation list                 ###
    ##################################################################
    my $self = shift;
    my $setup = shift;
    my $potlist = shift;
    return (0,0);
}

sub start_chrono()
{
    ######################################################
    ### start_chrono()                                 ###
    ### starts a chrono amperometry/potentiometry run  ###
    ######################################################
    return undef;
}

sub setup_chrono()
{
    ####################################################
    ### setup_crono()                                ###
    ### setup a chrono amperometry/potentiometry run ###
    ### with the specified polarisation and based on ###
    ### setup specified in hash ref                  ###
    ##################################################################
    ### If the setup key 'polarisation_offset' is set (and is      ###
    ### a function pointer) the function is called and the result  ###
    ### is added to the polarisation                               ###
    ##################################################################
    my $self = shift;
    my $pol = shift;
    my $setup = shift;
    return 0;
}

sub run_dvm()
{
    ########################################
    ### run_dvm()                        ###
    ### Returns a single dvm measurement ###
    ### Default returns "0.0000 0.0000"  ###
    ### the specific prosess value the   ###
    ### numbers represent depends on the ###
    ### device setup!                    ###
    ########################################
    my $self = shift;
    my $string = sprintf("%f %f",0,0);
    push @{$self->{'_data'}},$string;
    return $string;
}

sub set_resistor($)
{
    ########################################
    ### set_resistor($)                  ###
    ### Set the shunt ressitor value     ###
    ### Returns the last set value for   ###
    ### the internal current shunt       ###
    ### resistor, default is 1           ###
    ########################################
    my $self = shift;
    $self->{'_resistor'} = shift;
    return $self->{'_resistor'};
}

sub max_sweep_points()
{
    ############################################
    ### Max sweep points returns the maximum ###
    ### independent potential points for     ###
    ### a potential sweep                    ###
    ############################################
    my $self = shift;
    return $self->{'_max_sweep_points'};
}

return 1;
