package RemoteExec;
#$Id: RemoteExec.pm,v 1.9 2015/02/02 07:42:35 sofc Exp $

################################################################################################################ 
## 
## 
##
## 
##          Package RemoteExec
## 
## 
##
##  This package is intended to be used to manage execution of remote program from within Elchemea.
##  To obtain a remoteExec instance call the constructor with the filename and arguments of the program
##  in question (Remember to use the full path! as the module internally uses the path of '' intentionally):
##
##  $id = RemoteExec->new($command);
##
##  The constructor does some simple sanitizing, but not enough to ensure that no maliscious programs can not
##  be executed! Thus it is up to the user of RemoteExex to ensure that programs executed thorugh this interface
##  does not harm the system!
##  The constructor splits the parsed command into a filename and arguments and keeps those stored separately
##  making implementation of the filename and args functions simple.
##
##  Each RemoteExec instance has a number of member functions apart from those derived from Debug.pm:
##
##  $id->filename()       Returns the filename.
##  $id->args()           Returns a list of the arguments found as part of the original command 
##  $id->exists()         Returns true if the filename exists (result of the -e operator)
##  $id->exec(@)          Ececutes the program with the original arguments followed with the specified
##                        arguments (if any).
##  $id->exec_id($)       Same as exec, but without any extra program arguments, however the single argument to
##                        this function must be a instance of the FDEV class (or one of the derived classes).
##                        This member function is the only one directly linking the RemoteExec class to
##                        the Elchemea framework, so this class can be used without Elchemea as long as this
##                        member function is not used (as the RemoteExec class does not import any Elchemea
##                        classes apart from the Debug class (which is a class which only derives from Perl
##                        internal classes)
##
##
################################################################################################################

use strict;
use warnings;
use Debug;
our @ISA = qw(Debug);

BEGIN {
    use Exporter ();
    our ($VERSION, @ISA, @EXPORT, @EXPORT_OK, %EXPORT_TAGS);
    $VERSION = 1.0;
    %EXPORT_TAGS = (
        'standard'=>[],
                    );
    Exporter::export_ok_tags('standard');
    Exporter::export_tags('standard');
    @ISA = qw(Exporter);
}

sub new
{
    my $class = shift;
    my $self = $class->SUPER::new();
    my $command = shift;
    $command = '' unless ($command);
    $command =~ s/\n//g;
    my @args = split(/[\s\t]+/,$command);
    $self->{'_filename'} = shift(@args);
    $self->{'_args'} = \@args;
    return $self;
}

sub filename()
{
    my $self = shift;
    return $self->{'_filename'};
}

sub args()
{
    my $self = shift;
    return @{$self->{'_args'}};
}

sub exists()
{
    my $self = shift;
    return -e $self->{'_filename'};
}

sub exec
{
    my $self = shift;
    return undef unless ($self->filename && $self->exists);
    my $org_path = $ENV{'PATH'};
    $ENV{'PATH'} = '';
    my @args = $self->args;
    while (@_)
    {
	push @args,shift;
    }
    my $cmdstr = $self->filename.' '.join(' ',@args);
    $self->writedebug("Executing external command '$cmdstr'") if ($self->debug > 2);
    my @res = `$cmdstr`;
    foreach (@res)
    {
	chomp $_;
    }
    $self->writedebug("Result of external command '$cmdstr': ".join("\n",@res)) if ($self->debug);
    ### Restore original ENV;
    $ENV{'PATH'} = $org_path;
    return @res;
}

###################################################################################################
### exec_id is only used by the Elchemea system for impedance and electrochemical measurements ####
### Note that it inverts the result before returning the value !!!                             ####
###################################################################################################

sub exec_id
{
    my $self = shift;
    return undef unless ($self->filename);
    my $id = undef;
    $id = shift if (@_);
    my @args = $self->args;
    my $ext_pol_cmd = $self->filename.' '.join(' ',@args);
    $id->writedebug("Excuting '$ext_pol_cmd'") if (defined($id));
    my @res = $self->exec;
    my $result = shift @res; 
    my $last_message = undef;
    $last_message = join("\n",@res);
    eval {
	Elchemea::append_last_message($last_message);
    };
    if ($@)
    {
	$self->writedebug($@,1);
    }
    $id->writedebug("Result of '$ext_pol_cmd': $result") if (defined($id));
    $id->writedebug("Additional information from '$ext_pol_cmd': ".$last_message) if (defined($result) && $last_message && defined($id));
    if ($result =~ /([\d\.eE\-\+]+)/)
    {
	$result = $1;
    }
    else
    {
	$result = 0;
    }
    #### Return the negative value due to working electrode on device ground ####          
    return -$result;
}
